# DomWrapper wraps xml.dom to provide extra features that HiperDOM requires.
# HiperDom uses a special Text Node (AlreadyRendered) that can contain already
# rendered XML markup, and as such should not be subjected to the usual
# character data translations.
#
# Also provides wrapper conversion methods from xml to dom and vice versa, so
# that in the event of a DOM Level 3 compliant API for such conversions is
# released for Python we can switch easily.

from xml.dom.Text import Text
from xml.dom.ext.reader import PyExpat
from xml.dom import XML_NAMESPACE
from xml.dom.html import HTML_FORBIDDEN_END
from xml.dom.ext import Printer, SeekNss, GetAllNs
from string import find, strip
import re

try:
    from cStringIO import StringIO
except ImportError:
    from StringIO import StringIO

Reader=PyExpat.Reader

# Le Sigh. SAX2 doesn't do comments (SAX2-ext does, but python doesn't support
# that yet). And the PyExpat reader is seriously broken. Patchup time!
# All this *should* just work with future versions of 4Suite and PyXML.
if not hasattr(PyExpat, 'pyexpat'):
    try:
        # Python 2.0
        import pyexpat
    except ImportError:
        # Python 1.x with PyXML
        from xml.parsers import pyexpat

    PyExpat.pyexpat = pyexpat

    from xml.dom import DOMException
    class FtDomException(DOMException):
        pass

    XML_PARSE_ERROR = 1001

    def fixed_initState(self, ownerDoc=None):
        self._ownerDoc = None
        self._rootNode = None
        #Set up the stack which keeps track of the nesting of DOM nodes.
        self._nodeStack = []
        if ownerDoc:
            self._ownerDoc = ownerDoc
            #Create a docfrag to hold all the generated nodes.
            self._rootNode = self._ownerDoc.createDocumentFragment()
            self._nodeStack.append(self._rootNode)
        self._dt = None
        self._xmlDecl = None
        self._orphanedNodes = []
        self._namespaces = {'xml': XML_NAMESPACE}
        self._namespaceStack = []
        self._currText = ''
        return
else:
    from xml.dom.ext import FtDomException, XML_PARSE_ERROR

Reader.initState = fixed_initState

def todom(xml, _WS=re.compile('[ \t\r\n]+'),
    _extId=re.compile('(?:\'([^\']*)\'|"([^"]*)")[ \t\r\n]+'\
                      '(?:\'([^\']*)\'|"([^"]*)")')):
    """Convert an XML document to a DOM tree."""

    reader = Reader()
    
    # We do our own 'fromStream' because we want the prolog information (DTDs
    # and such) to be in the DOM as well. PyExpat skips to the first element.
    reader.initParser()
    reader.initState()

    # Now we parse out the prolog (minus the XMLDecl)
    # We don't check for correct XML, that's something the parser will do for
    # us. We are looking for the DTD, because PyExpat doesn't do that for us.
    # Expat 1.2 *does* (and the reader has support for it), but PyExpat hasn't
    # been updated yet. *Grmbl*
    docTypeFound = 0
    start = find(xml, '<?xml')
    end = find(xml, '?>', start)
    if start > -1 and end > start:
        start = end + 2
    else:
        start = 0

    while 1:
        tagStart = find(xml, '<', start)

        # Opening element. Let PyExpat take it from here
        if xml[tagStart + 1] not in ('!', '?'):
            break

        # A comment, skip it.
        if xml[tagStart:tagStart + 4] == '<!--':
            tagEnd = find(xml, '-->', tagStart + 4)
            if tagEnd == -1:
                break # Let PyExpat work out the error
            start = tagEnd + 3
            continue

        # A Processing Instruction, skip it.
        if xml[tagStart:tagStart + 2] == '<?':
            tagEnd = find(xml, '?>', tagStart + 2)
            if tagEnd == -1:
                break # Let PyExpat work out the error
            start = tagEnd + 2
            continue

        # A Document Type Declaration
        # We are not interested in any actual markup declarations,
        # only in a SYSTEM and PUBLIC ID
        # I know, DOM level 2 *does* support entities and notations.
        # But this is meant to be kept simple, and it will suffice for
        # 99.99% of all XHTML documents and most XML as well.
        # Get PyExpat updated to Expat 1.2 and we'll talk (or rather,
        # PyExpat.Reader will).
        if xml[tagStart:tagStart + 9] == '<!DOCTYPE':
            markupStart = find(xml, '[', tagStart)
            if markupStart > -1:
                # Skip the markup declaration
                tagEnd = find(xml, ']', markupStart)
                if tagEnd > -1:
                    tagEnd = find(xml, '>', tagEnd)
            else:
                tagEnd = find(xml, '>', tagStart)
                markupStart = tagEnd
                
            if tagEnd > -1:
                if docTypeFound: break
            
                doctype = strip(xml[tagStart+9:tagEnd])
                nameEnd = _WS.search(doctype)
                if not nameEnd: break
                else: nameEnd = nameEnd.start()

                externalId = strip(
                    doctype[nameEnd:markupStart-(tagStart+9)])
                if externalId[:6] == 'SYSTEM':
                    public = ''
                    system = strip(externalId[7:])[1:-1]
                elif externalId[:6] == 'PUBLIC':
                    result = _extId.match(strip(externalId[7:]))
                    if not result: break 

                    result = result.groups()
                    public = result[0] or result[1] # ' or "
                    system = result[2] or result[3] # ' or "
                else:
                    public = ''
                    system = ''

                reader.startDTD(strip(doctype[:nameEnd]),
                    public, system)
                    
            else:
                break # Let PyExpat work out the error
            start = tagEnd + 1
            docTypeFound = 1
            continue
        
    # Now let PyExpat parse the document or declare any errors
    stream = StringIO(xml)
    success = reader.parser.ParseFile(stream)
    if not success:
        if self._rootNode: PyExpat.ReleaseNode(reader._rootNode)
        if self._ownerDoc: PyExpat.ReleaseNode(reader._ownerDoc)
        raise FtDomException(XML_PARSE_ERROR, (
            reader.parser.ErrorLineNumber, reader.parser.ErrorColumnNumber,
            PyExpat.pyexpat.ErrorString(reader.parser.ErrorCode)))
    reader._completeTextNode()

    rootNode = reader._rootNode or reader._ownerDoc
    
    stream.close()
    
    return rootNode

def toxml(dom):
    """Convert a DOM tree to a XML tree
    
    Converts a xml.dom tree back to an xml string, but with special
    consideration for HiperDom.
    
    """
    
    stream = StringIO()
    nss_hints = SeekNss(dom)
    visitor = HdomPrintVisitor(stream, 'UTF-8', nss_hints)
    Printer.PrintWalker(visitor, dom).run()

    stream.seek(0)
    return stream.read()


class AlreadyRendered(Text):
    """Special Text Node subclass for HiperDom purposes.
    
    Special Text Node subclass that will not have the usual markup escaping
    applied. It is used in stead of Text nodes there were already rendered XML
    markup is returned.
    
    Since only the fact that we will test for isinstance(instance,
    AlreadyRendered) no actual new methods are declared.

    """

    pass


# We provide or own visitor class, because #1 we want AlreadyRendered nodes to
# be treated differently, #2 PyXML 0.6.2 DTD printing is horked, #3 we want to
# output HTML documents empty elements printed in a old-browser-friendly
# fashion and #4, we want the prolog to stay readable, so we add whitespace
# after an XML declaration and a DTD.
class HdomPrintVisitor(Printer.PrintVisitor):
    """Modified PrinterVisitor for special HiperDom AlreadyRendered tricks."""

    def visit(self, node):
        if isinstance(node, AlreadyRendered):
            return self.stream.write(Printer.TranslateCdata(node.data,
                self.encoding, markupSafe=1)) 

        else: 
            return Printer.PrintVisitor.visit(self, node)

    def visitDocumentType(self, node):
        if not node.entities.length and not node.notations.length and \
            not node.systemId:
            return

        self.stream.write("<!DOCTYPE " + node.name)
        self.__emptyReturn = 0

        if node.systemId != '':
            if node.publicId != '':
                if '"' not in node.publicId:
                    self.stream.write(' PUBLIC "' + node.publicId + '" ')
                else:
                    self.stream.write(" PUBLIC '" + node.publicId + "' ")
            else:
                self.stream.write(' SYSTEM ')

            if '"' not in node.systemId:
                self.stream.write('"' + node.systemId + '"')
            else:
                self.stream.write("'" + node.systemId + "'")

        if node.entities.length or node.notations.length:
            self.stream.write(' [')
            self.visitNamedNodeMap(node.entities)
            self.visitNamedNodeMap(node.notations)
            self.stream.write(']')
        self.stream.write('>\n')

        return

    def visitElement(self, node):
        #FIXME: Borrow opt from Sax2
        self._namespaces.append(self._namespaces[-1].copy())
        self.stream.write('<' + node.tagName)
        if not hasattr(node.ownerDocument,'isXml') or node.ownerDocument.isXml():
            st = ''
            nss = GetAllNs(node)
            if self._nsHints:
                self._nsHints.update(nss)
                nss = self._nsHints
                self._nsHints = {}
            del nss['xml']
            for prefix in nss.keys():
                if not self._namespaces[-1].has_key(prefix) or self._namespaces[-1][prefix] != nss[prefix]:
                    if prefix:
                        st = st + ' xmlns:' + prefix + "='" + nss[prefix] + "'"
                    else:
                        st = st + ' xmlns' + "='" + nss[prefix] + "'"
                self._namespaces[-1][prefix] = nss[prefix]
            self.stream.write(st)
        self.visitNamedNodeMap(node.attributes)
        if len(node.childNodes):
            self.stream.write('>')
            self.visitNodeList(node.childNodes)
            if not hasattr(node.ownerDocument,'isXml') or node.ownerDocument.isXml() or (node.tagName not in HTML_FORBIDDEN_END):
                self.stream.write('</' + node.tagName + '>')
        elif node.ownerDocument.doctype.name == 'html':
            # Old-browser friendly XHTML output
            self.stream.write(' />')
        elif not hasattr(node.ownerDocument,'isXml') or node.ownerDocument.isXml():
            self.stream.write('/>')
        elif node.tagName not in HTML_FORBIDDEN_END:
            self.stream.write('></' + node.tagName + '>')
        else:
            self.stream.write('>')
        del self._namespaces[-1]
        return

    def visitProlog(self):
        Printer.PrintVisitor.visitProlog(self)
        self.stream.write('\n')
