__doc__ = 'Strategy objects to process HiperDom directives'

from xml.dom.Node import Node
from DomWrapper import AlreadyRendered, toxml
from DocumentTemplate.DT_Util import InstanceDict
from utils import getObject

# Namespace URI for HiperDom attributes
HDOM_NAMESPACE = 'uri:hiperdom'

class _strategy:
    def __init__ (self, template, context):
        self.template = template
        self.context = context

class assetHunter (_strategy):
    def __init__ (self, parent, preserveDirectives=0):
        self.preserveDirectives = preserveDirectives
        self.parent = parent
        _strategy.__init__(self, parent.template, parent.context)

    def do_id (self, element, outDoc, specialData):
        """Process the data and add the results to specialData with
        the name specified in the hdom:id directive
        """
        name = element.getAttributeNS (HDOM_NAMESPACE, 'id')
        wkcopy = outDoc.importNode(element, 1)
        # We don't want to remove the hdom:id, probably because we're running
        # from inside a textExpander
        if not self.preserveDirectives:
            # remove the hdom:id attribute
            wkcopy.removeAttributeNS (HDOM_NAMESPACE, 'id')
        specialData[name] = toxml(self.template._processElement(wkcopy, 
            self.parent, outDoc))
        return wkcopy

class textExpander (_strategy):
    def __init__ (self, template, context, preserveDirectives=0):
        self.preserveDirectives = preserveDirectives
        _strategy.__init__(self, template, context)
        
    def do_text (self, element, outDoc, specialData):
        """Replace the contents (childNodes) of the element by the
        result of evaluating the identifier passed as the value of
        the hdom:text directive.
        But first, hunt for any assets in the subtree to be discarded
        (assets are elements with an hdom:id attribute); the assets
        are added to the namespace when the identifier is evaluated.
        Finally, here we don't want to strip the Hdom directives
        unless explicitly told so
        """
        name = element.getAttributeNS (HDOM_NAMESPACE, 'text')
        wkcopy = outDoc.importNode(element, 0)
        assets = {}
        if not self.preserveDirectives:
            wkcopy.removeAttributeNS (HDOM_NAMESPACE, 'text')
        self.template._processElement(element, 
            assetHunter(self, self.preserveDirectives), outDoc, assets)

        md = self.context
        md._push(assets)
        try:
            # Asset retrieval is now only available for DTML and PythonScripts.
            value = getObject(name, md, self.template)
        finally:
            md._pop()
        
        wkcopy.appendChild(AlreadyRendered(outDoc, str(value)))
        return wkcopy

    def do_include(self, element, outDoc, specialData):
        """Includes the DOM of another HiperDOm template into this one.

        For now, this only replaces the element with the included DOM from the
        other HiperDom template. More functionality is planned.

        """

        name = element.getAttributeNS (HDOM_NAMESPACE, 'include')

        assets = {}
        self.template._processElement(element, 
            assetHunter(self, self.preserveDirectives), outDoc, assets)
        
        md = self.context
        md._push(assets)
        try:
            template = getObject(name, md, self.template, call=0)
        finally:
            md._pop()
        
        if (type(template) is not type(self.template) or
            not hasattr(template, 'isXmlTemp')):
            raise TypeError, '%s is not a HiperDom template' % name

        # Check for element name match
        if template._dom.doctype.name != element.nodeName:
            raise ValueError, 'Cannot include mismatching document type '\
              'from template %s' % name

        # Empty DOM?
        if not template._dom.documentElement:
            # if expanding a template, return empty original element
            if self.preserveDirectives:
                return outDoc.importNode(element, 0)
            else:
                return None

        # Clone and process the foreign template
        wkcopy = self.template._processElement(
            outDoc.importNode(template._dom.documentElement, 1), self, outDoc,
                specialData)
        
        # Expansion of this template, preserve attributes
        if self.preserveDirectives:
            wkcopy.setAttributeNodeNS(outDoc.importNode(
                element.getAttributeNodeNS(HDOM_NAMESPACE, 'include'), 0))
        
        return wkcopy

class fullHdom (textExpander):
    def do_dummy (self, element, outDoc, specialData):
        """Discard this element"""
        return None

    def do_condition (self, element, outDoc, specialData):
        """Evaluate the name given as the directive's value, then leave the
        contents untouched if the name evaluates to true and discard the
        element if it evaluates to false.
        """
        name = element.getAttributeNS (HDOM_NAMESPACE, 'condition')

        if getObject(name, self.context, self.template):
            wkcopy = outDoc.importNode(element, 1)
            wkcopy.removeAttributeNS (HDOM_NAMESPACE, 'condition')
            return wkcopy
        else:
            return None

    def do_loop (self, element, outDoc, specialData):
        """Evaluate the name given as the directive's value, then process
        the contents of the element once for every item of the result.
        It is an error for the result not to be a sequence object.
        """
        if element.ownerDocument.documentElement is element:
            raise 'TemplateError', 'Loop attribute cannot be set on ' \
                'root element'
        
        name = element.getAttributeNS (HDOM_NAMESPACE, 'loop')
        fragment = outDoc.createDocumentFragment()
        namespace = self.context

        for item in getObject(name, namespace, self.template):
            namespace._push({'sequence_item': item})
            namespace._push(InstanceDict(item, namespace))
            newStrategy = fullHdom(self.template, namespace)
            
            wkcopy = outDoc.importNode(element, 1)
            wkcopy.removeAttributeNS (HDOM_NAMESPACE, 'loop')
            try:
                child = self.template._processElement(wkcopy, newStrategy, 
                    outDoc)
            finally:
                namespace._pop(2)

            fragment.appendChild(child)
        
        return fragment

    def do_method (self, element, outDoc, specialData):
        """Replace the whole element by a new element resulting from
        evaluating the identifier passed as the value of the hdom:method
        directive.
        The method will be called with a special parameter: element_data,
        which is the xml.dom.Element object.        
        It may return a dictionary with some special conventions (keys are
        attributes, a special key '<text>' is the content), or a string

        or a new xml.dom Element object.
        """
        name = element.getAttributeNS (HDOM_NAMESPACE, 'method')
        wkcopy = outDoc.importNode(element, 1)
        wkcopy.removeAttributeNS (HDOM_NAMESPACE, 'method')
        wkcopy.__allow_access_to_unprotected_subobjects = 1

        md = self.context
        md._push({'element_data': wkcopy})
        try:
            result = getObject(name, md, self.template)
        finally:
            md._pop()

        if result is None or hasattr(result, 'nodeType'):
            return result
        elif type(result) == type(''):
            return AlreadyRendered(outDoc, result)
        elif type (result) == type ({}):
            newtext = ''
            if result.has_key ('<text>'):
                newtext = result['<text>']
                del result['<text>']
            for name in wkcopy.attributes.keys():
                wkcopy.removeAttribute(name)
            for name, value in result.items():
                wkcopy.setAttributeNS('', name, value)
            for child in wkcopy.childNodes[:]:
                wkcopy.removeChild(child)
            if newtext:
                wkcopy.appendChild(AlreadyRendered(outDoc, newtext))
            return wkcopy
        else:
            raise TypeError, 'invalid object returned from hdom:method'

    def do_call(self, element, outDoc, specialData):
        """Call the named object, but don't alter the output DOM in any way.

        Very handy for making side-effects happen, like settng cookies and other
        headers.

        """

        name = element.getAttributeNS (HDOM_NAMESPACE, 'call')
        wkcopy = outDoc.importNode(element, 0)
        wkcopy.removeAttributeNS (HDOM_NAMESPACE, 'call')

        getObject(name, self.context, self.template)

        # child nodes are processed just the same way.
        for child in element.childNodes:
            newChild = self.template._processElement(child, self, outDoc, 
                specialData)
            if newChild:
                wkcopy.appendChild(newChild)

        return wkcopy
