__doc__ = """
HiperDom Template objects
HiperDom Templates are XML page templates - meaning, they can generate XML
documents if they're processed in a Zope (for example) context. The primary
use for them is generating xHTML webpages.

HiperDom works by recognizing special XML attributes (called Directives),
marked with a "hdom:" XML namespace. For example,
<title hdom:text="docTitle">foo</title> will replace "foo" by the value
resulting from evaluating the variable, property or function "docTitle".
"""
__version__ = '0.4.0'

import Globals, Acquisition, AccessControl.Role, OFS.SimpleItem
import OFS.PropertyManager, DateTime
import string
import HiperDomStrategies, DomWrapper, utils
from xml.dom.Node import Node
from DocumentTemplate.DT_Util import InstanceDict, TemplateDict

_hdom_deprecated_attrs = ()
_hdom_prototype = {'HTML': 0, 
        'text': "<?xml version='1.0' encoding='UTF-8'?>\n<EmptyTemplate xmlns:hdom='uri:hiperdom'/>",
        '_dom': None, 'title': 'HiperDom Template', 'content_type': '',
    }

HDOM_NAMESPACE = HiperDomStrategies.HDOM_NAMESPACE
SIMPLE_NODES = (Node.ATTRIBUTE_NODE, Node.TEXT_NODE, Node.TEXT_NODE,
    Node.CDATA_SECTION_NODE, Node.ENTITY_NODE, Node.PROCESSING_INSTRUCTION_NODE,
    Node.COMMENT_NODE, Node, Node.NOTATION_NODE)

class HiperDomTemplate (
    OFS.SimpleItem.Item,
    Globals.Persistent,
    Acquisition.Implicit,
    AccessControl.Role.RoleManager,
    OFS.PropertyManager.PropertyManager
    ):
    """HiperDom Template class
    __call__ and __str__ render the template;
    manage_getTemplate returns the source, and sets content-type;
    text_expand processes only hdom:text directives, leaving them in place and
        respecting hdom:id assets
    The template object is "meaningful" in the namespace, or in other words,
        it's parallel to a DTML Document, not a DTML Method
    """

    #################################################
    ### Zope class attributes

    name = 'HiperDomTemplate'
    id = 'HiperDomTemplate'
    meta_type = 'HiperDom Template'
    icon = 'misc_/HiperDom/HiperDomTemplate.gif'
    # Tell DocumentTemplate that we accept its calling conventions...
    isDocTemp = 1
    # Identify this as a Xml template.
    isXmlTemp = 1

    index_html = None # Prevent accidental acquisition

    # We pretend to be a function, just like DTML* objects
    class func_code: pass
    func_code.__roles__=()
    func_code=func_code()
    func_code.co_varnames='self','REQUEST','RESPONSE'
    func_code.co_argcount=3

    func_defaults__roles__=()
    func_defaults=()

    _properties = (
            {'id': 'HTML', 'type': 'boolean', 'mode': 'w'},
            {'id': 'title', 'type': 'string', 'mode': 'w'},
            {'id': 'content_type', 'type': 'string', 'mode': 'w'},
        )

    manage_options = (
            {'label':'Edit', 'action':'manage_main',
             'help': ('HiperDom', 'TemplateEdit.stx')},
            {'label':'Expand', 'action':'manage_expand',
             'help': ('HiperDom', 'TemplateExpand.stx')},
            {'label':'View', 'action':''},
            {'label':'Properties', 'action':'manage_propertiesForm',
             'help': ('OFSP', 'Properties.stx')},
            {'label':'Security', 'action':'manage_access',
             'help': ('OFSP', 'Security.stx')},
            {'label':'Ownership', 'action':'manage_owner',
             'help': ('OFSP', 'Ownership.stx')},
        )


    __ac_permissions__ = (
            ('View management screens',
             ('manage_tabs', 'manage_main', 'manage_expand',
              'manage_getTemplate', 'text_expand', 'document_src')),
            ('Change Permissions',
             ('manage_access',)),
            ('Change HiperDom Templates',
             ('manage_edit', 'PUT')),
            ('Manage Properties',
             ('manage_addProperty', 'manage_editProperties',
             'manage_delProperties', 'manage_changeProperties',)),
            ('View',
             ('', '__call__',)),
            ('FTP access',
             ('manage_FTPstat','manage_FTPget','manage_FTPlist')),
            ('Know the secrets of life',
             ('manage_dict',)),
            )


    #################################################
    ## Instance constructors

    def __init__ (self, id):
        "Initialize the class instance"
        self.id = id
        for name, value in _hdom_prototype.items ():
            if not hasattr (self, name):
                setattr (self, name, value)
        if not self._dom:
            self._updateDom()

    def __setstate__ (self, state):
        "Initialize when loading from the ZODB"
        Globals.Persistent.__setstate__ (self, state)
        for name, value in _hdom_prototype.items ():
            if not hasattr (self, name):
                setattr (self, name, value)
        for name in _hdom_deprecated_attrs:
            if hasattr (self, name):
                delattr (self, name)
        if not self._dom or not self._dom.documentElement:
            self._updateDom()


    #################################################
    ## This does the "Narower/Wider/Taller/Shorter"
    ## manage_main textarea magic
    ## (shamelessly copied from OFS.DTMLMethod)

    _size_changes = {
        'Bigger': (5,5),
        'Smaller': (-5,-5),
        'Narrower': (0,-5),
        'Wider': (0,5),
        'Taller': (5,0),
        'Shorter': (-5,0),
    }

    def _er (self, text, submit, dtpref_cols, dtpref_rows, REQUEST):
        dr, dc = self._size_changes[submit]
        
        rows = max (1, string.atoi (dtpref_rows) + dr)
        cols = max (40, string.atoi (dtpref_cols) + dc)
        e = (DateTime.DateTime ('GMT') + 365).rfc822()
        resp = REQUEST['RESPONSE']
        resp.setCookie ('dtpref_rows', str (rows), path='/', expires=e)
        resp.setCookie ('dtpref_cols', str (cols), path='/', expires=e)
        return self.manage_main (self, REQUEST, text=text, dtpref_cols=cols, dtpref_rows=rows)


    #################################################
    ## Management interface

    manage_main = Globals.HTMLFile ('HiperDomTemplateEdit', globals())
    manage_expand = Globals.HTMLFile ('HiperDomTemplateExpand', globals())

    def manage_edit (self, text, submit='Change', dtpref_cols='50',
                    dtpref_rows='20', REQUEST=None):
        "Edit the template."
        if self._size_changes.has_key (submit):
             return self._er (text, submit, dtpref_cols, dtpref_rows, REQUEST)
        
        self.text = text
        self._updateDom()

        if REQUEST is not None:
            return self.manage_main (self, REQUEST,
                manage_tabs_message = 'Template changed.')


    #################################################
    ## Externally accessible, but not from the manage
    ## interface
    
    def manage_getTemplate (self, RESPONSE=None):
        """Return the template.

        It is rendered from the dom tree instead of just self.text, so that we
        return a normalized version. The template is expanded if possible.

        """
        if RESPONSE is not None:
            self._setContentType(RESPONSE)
        return str(self)
    
    # this alias makes WebDAV happy
    # (it allows you to use the "source link property" to
    # get the source instead of the rendered template, when
    # you're retrieving the document for edition via DAV)
    document_src = manage_getTemplate

    # This alias is to make FTP get possible.
    manage_FTPget = manage_getTemplate

    def PUT (self, REQUEST, RESPONSE):
        "Change the template from an HTTP PUT"
        self.dav__init (REQUEST, RESPONSE)
        self.text = REQUEST.get ('BODY', '')
        self._updateDom()
        RESPONSE.setStatus(204)
        return RESPONSE


    #################################################
    ## Externally accessible from Python/DTML only
    
    # Security related (?)
    validate = None
    validate__roles__ = ()
    
    def __call__ (self, client=None, REQUEST=None, RESPONSE=None, **kw):
        "Render the template"
        if RESPONSE is not None:
            self._setContentType(RESPONSE)

        bself = getattr(self, 'aq_explicit', self)
        if client is None:
            # We are the client. Note that we already fool the Publisher into
            # giving us ourselves as the client parameter with all that
            # 'func_code' magic earlier.
            client = bself
        else:
            # Called as a subtemplate, but we want to participate in the context
            # as well.
            client = (client, bself)

        # Now build a TemplateDict if necessary. This will be our context.
        pushed=None
        try:
            if REQUEST.__class__ is TemplateDict: pushed=0
        except: pass

        if pushed is not None:
            # We were passed a TemplateDict, so we must be a sub-template
            md=REQUEST
            push=md._push
        else:
            md=TemplateDict()
            push=md._push
            if REQUEST:
                push(REQUEST)
            md.validate=self.validate
            if type(client)==type(()):
                md.this=client[-1]
            else: md.this=client
            pushed=0

        # We participate nicely, so we also play along with the levels.
        level=md.level
        if level > 200: raise SystemError, (
            'infinite recursion in document template')
        md.level=level+1

        if client is not None:
            if type(client)==type(()):
                # if client is a tuple, it represents a "path" of clients
                # which should be pushed onto the md in order.
                for ob in client:
                    push(InstanceDict(ob, md)) # Circ. Ref. 8-|
                    pushed=pushed+1
            else:
                # otherwise its just a normal client object.
                push(InstanceDict(client, md)) # Circ. Ref. 8-|
                pushed=pushed+1
                
        if kw:
            push(kw)
            pushed=pushed+1

        try:
            return DomWrapper.toxml(self._processElement(self._dom, 
                HiperDomStrategies.fullHdom(self, md)))
        finally:
            if pushed: md._pop(pushed) # Get rid of circular reference!
            md.level=level # Restore previous level

    def __str__(self):
        # return the source; expanded if possible, otherwise normalised.
        try:
            return self.text_expand()
        except:
            return DomWrapper.toxml(self._dom)

    def text_expand (self, REQUEST=None):
        """Process only hdom:text directives, leaving them in place and
        respecting hdom:id assets
        """

        if REQUEST is None:
            REQUEST = self.REQUEST

        md = TemplateDict()
        md.validate = self.validate
        md.this = self
        md._push(REQUEST)
        md._push(InstanceDict(self, md))
        try: 
            return DomWrapper.toxml(self._processElement(self._dom, 
                HiperDomStrategies.textExpander(self, md, 
                    preserveDirectives=1)))
        finally:
            md._pop()

    #################################################
    ## Methods for internal consumption
    
    def _updateDom (self):
        "Update the internal DOM tree from the text"
        self._dom = DomWrapper.todom(self.text)
    
    def _setContentType (self, RESPONSE):
        "Set the content-type"
        # FIXME: we want more robustness, like checking if the
        # first element in the DOM is an <html> and if we have a
        # Content-type property
        have_key=RESPONSE.headers.has_key
        if not (have_key('content-type') or have_key('Content-Type')):
            if self.content_type:
                ct = self.content_type
                if string.find(ct, 'charset') == -1:
                    ct = '%s;charset=UTF-8' % ct
                RESPONSE.setHeader('Content-type', ct)
                return

            if self._dom.doctype.name == 'html':
                RESPONSE.setHeader('Content-type', 'text/html;charset=UTF-8')

        return # default content-type handling

    default_content_type = 'text/xml;charset=UTF-8'
        
    def _processElement (self, element, strategy, outDoc=None, specialData={}):
        """hdom:process one element (and its children) using this strategy
        returns the new, processed Node object (or an DomWrapper.AlreadyRendered
        pseudo-Node if the strategy returned a string)
        specialData may be used to "fish" for data in the tree; for
        example, it is used to gather assets marked with hdom:id
        before gobbling an hdom:text element
        """

        # Special case: the document element should be copied to form the basis
        # of the output DOM. It is then subsequently used to import all rendered
        # elements.
        if element.nodeType == element.DOCUMENT_NODE:
            if element.doctype is not None: 
                docType = element.doctype.cloneNode(1)
            else:
                docType = None 
            outDoc = element.implementation.createDocument(None, None, docType)
            
            # Clone any non-document children of the template DOM; they can't be
            # elements and thus are not affected by the rendering process.
            for child in element.childNodes: 
                if child.nodeType == child.DOCUMENT_TYPE_NODE: continue
                if child == element.documentElement: continue
                outDoc.appendChild(outDoc.importNode(child, 0))

            if element.documentElement:
                newDocEl = self._processElement(element.documentElement,
                    strategy, outDoc, specialData)
                if newDocEl:
                    outDoc.appendChild(newDocEl)

            return outDoc
        
        assert outDoc != None, 'No output document'
        
        # Skip nodes that don't have attributes anyway
        if element.nodeType in SIMPLE_NODES:
            return outDoc.importNode(element, 1)
        
        # handle hdom: attributes
        attrs = element.attributes
        if attrs:
            for key in filter(lambda k: k[0] == HDOM_NAMESPACE, attrs.keys()):
                if hasattr(strategy, 'do_' + key[1]):
                    wkcopy = getattr(strategy, 'do_' + key[1])(element,
                        outDoc, specialData)
                    if type (wkcopy) == type (''):
                        return DomWrapper.AlreadyRendered(outDoc, wkcopy)
                    else:
                        # We don't want to process the childNodes of already
                        # processed elements. Processed elements take care of
                        # that themselves. 
                        return wkcopy

        # we don't want to modify the upstream element object
        wkcopy = outDoc.importNode(element, 0)

        # process the children
        for child in element.childNodes:
            newChild = self._processElement(child, strategy, outDoc, 
                specialData)
            if newChild:
                wkcopy.appendChild(newChild)
        return wkcopy

    #################################################
    ## Debugging stuff, in the "what not even God
    ## should do" category
    def manage_dict (self, REQUEST=None):
        "Debugging method - meaning is obvious"
        return `self.__dict__`

Globals.default__class_init__(HiperDomTemplate)

manage_addHiperDomTemplateForm = Globals.HTMLFile ('HiperDomTemplateAdd', globals ())

def manage_addHiperDomTemplate (self, id, REQUEST=None):
    "Add an HiperDomTemplate to the current ObjectManager"
    obj = HiperDomTemplate (id)
    assert obj, "Could not create template!"
    obj.title = REQUEST['title']
    self._setObject (id, obj)
    if REQUEST is not None:
        return self.manage_main (self, REQUEST)
