__doc__ = """
ZUnit tests for the HiperDomTemplate class
"""

from Products import ZUnit
import HiperDomTemplate
import string

class creationTest (ZUnit.TestCase):
    def runTest (self):
        "Try to create a template"
        ctx = self.ZopeContext
        ctx.REQUEST.set ('title', 'Testing creation')
        HiperDomTemplate.manage_addHiperDomTemplate (ctx, 'created', ctx.REQUEST)
        self.failUnless ('created' in ctx.objectIds(), 'Couldn\'t create HiperDom Template')
        try:
            template = ctx.created.manage_getTemplate()
            self.failUnless (template and template[0] == '<' and template[-1] == '>', 'HiperDom Template not properly initialized')
        except:
            self.fail ('HiperDom Template not properly initialized, couldn\'t render source template:' + ZUnit.subTraceback())

class objectTests (ZUnit.TestCase):
    def setUp (self):
        ctx = self.ZopeContext
        self.r = ctx.REQUEST
        self.r.set ('title', 'Testing HiperDom')
        HiperDomTemplate.manage_addHiperDomTemplate (ctx, 'template', ctx.REQUEST)
        self.t = ctx.template

    def tearDown (self):
        self.ZopeContext.manage_renameObject ('template',
            'template%02d' % len (self.ZopeContext.objectIds()))

    def test01_manage_main (self):
        "Try to access manage_main"
        try:
            page = self.t.manage_main (self.t, self.r)
        except:
            self.fail ('Couldn\'t view management screen (manage_main):' + ZUnit.subTraceback())
        self.failUnless (string.find (page, ">&lt;?xml version='1.0' "
            "encoding='UTF-8'?&gt;\n&lt;EmptyTemplate "
            "xmlns:hdom='uri:hiperdom'/&gt;</TEXTAREA>") > 0,
            'Management screen corrupted (manage_main)')
    
    def test02_manage_edit (self):
        "Try to edit the template via manage_edit"
        try:
            self.t.manage_edit (text='<foo><bar name="Madame Satan" city="Sao Paulo"/></foo>')
        except:
            self.fail ('Couldn\'t edit template:' + ZUnit.subTraceback())
        t = self.t.manage_getTemplate()
        self.failUnless (t == "<?xml version='1.0' encoding='UTF-8'?>\n<foo>"
            "<bar name='Madame Satan' city='Sao Paulo'/></foo>",
            'Something creepy happened with my template:\n         ' + t)

    def makeRenderTest (self, substring, invert=None):
        "Try to render the template, then check if the results contain 'substring'"
        try:
            t = self.t()
        except:
            self.fail ('Couldn\'t render template:' + ZUnit.subTraceback())
        if invert:
            test = self.failIf
        else:
            test = self.failUnless
        test (string.find (t, substring) >= 0,
            'The template was rendered incorrectly:\n         ' +
            string.join (string.split (t, '\n'), '\n         '))
    
    def test03_renderSimpeTemplate (self):
        "Try to render a template that doesn't contain any HiperDom directives"
        self.test02_manage_edit()
        self.makeRenderTest ('Madame Satan')
    
    def test04_renderDummyDirective (self):
        "Try to use the hdom:dummy directive"
        self.t.manage_edit (text='<foo xmlns:hdom="uri:hiperdom"><bar hdom:dummy="1"/></foo>')
        self.makeRenderTest ('<bar', 1)
    
    def test05_renderTextDirective (self):
        "Try to use the hdom:text directive"
        self.t.manage_edit (text='<foo xmlns:hdom="uri:hiperdom" hdom:text="bar"></foo>')
        self.t.bar = 'Madame Satan'
        
        self.makeRenderTest ("<?xml version='1.0' encoding='UTF-8'?>\n<foo xmlns:hdom='uri:hiperdom'>Madame Satan</foo>")
    
    def test06_renderTrueConditionDirective (self):
        "Try to use the hdom:condition directive with a true condition"
        self.t.manage_edit (text='<foo xmlns:hdom="uri:hiperdom"><bar hdom:condition="yes"/></foo>')
        self.t.yes = 1
        self.makeRenderTest ('<bar')
    
    def test07_renderFalseConditionDirective (self):
        "Try to use the hdom:condition directive with a false condition"
        self.t.manage_edit (text='<foo xmlns:hdom="uri:hiperdom"><bar hdom:condition="no"/></foo>')
        self.t.no = 0
        self.makeRenderTest ('<bar', 1)
    
    def test08_renderLoopDirective (self):
        "Try to use the hdom:loop directive to loop over a list of 3 ints"
        self.t.manage_edit (text='<foo xmlns:hdom="uri:hiperdom"><baz hdom:loop="list"><bar hdom:text="sequence_item"/></baz></foo>')
        self.t.list = [1, 2, 3]
        try:
            t = self.t()
        except:
            self.fail ('Couldn\'t render template:' + ZUnit.subTraceback())
        self.failUnless ((string.find (t, '<bar>1') >= 0) and
            (string.find (t, '<bar>2') >= 0) and (string.find (t, '<bar>3') >= 0),
            'The template was rendered incorrectly:\n         ' +
            string.join (string.split (t, '\n'), '\n         '))
    
    def test09_renderMethodDirectiveReturningNone (self):
        "Try to use the hdom:method directive with a method that returns None"
        self.t.manage_edit (text='<foo xmlns:hdom="uri:hiperdom"><bar hdom:method="none"/></foo>')
        def return_None (element_data=None):
            assert element_data is not None, 'No element_data passed in!'
            return None
        self.t.none = return_None
        self.makeRenderTest ('<bar', 1)
    
    def test10_renderMethodDirectiveReturningString (self):
        "Try to use the hdom:method directive with a method that returns a string"
        self.t.manage_edit (text='<foo xmlns:hdom="uri:hiperdom"><bar hdom:method="somethod"/></foo>')
        def somethod (element_data=None):
            assert element_data is not None, 'No element_data passed in!'
            return 'It worked.'
        self.t.somethod = somethod
        self.makeRenderTest ('It worked.')
    
    def test11_renderMethodDirectiveReturningXML (self):
        "Try to use the hdom:method directive with a method that returns a xml node object"
        self.t.manage_edit (text='<foo xmlns:hdom="uri:hiperdom"><bar hdom:method="somethod"/></foo>')
        def somethod (element_data=None):
            assert element_data is not None, 'No element_data passed in!'
            return element_data.ownerDocument.createElement('success')
        self.t.somethod = somethod
        self.makeRenderTest ('<success')
        pass
    
    def test12_renderMethodDirectiveReturningDict (self):
        """Try to use the hdom:method directive with a method that returns a
        dictionary representing an empty element"""
        self.t.manage_edit (text='<foo xmlns:hdom="uri:hiperdom"><bar hdom:method="somethod"/></foo>')
        def somethod (element_data=None):
            assert element_data is not None, 'No element_data passed in!'
            return {'name': 'Madame Satan'}
        self.t.somethod = somethod
        self.makeRenderTest ("<bar name='Madame Satan'/>")
    
    def test13_renderMethodDirectiveReturningDictWithText (self):
        """Try to use the hdom:method directive with a method that returns a
        dictionary representing an element containing text"""
        self.t.manage_edit (text='<foo xmlns:hdom="uri:hiperdom"><bar hdom:method="somethod"/></foo>')
        def somethod (element_data=None):
            assert element_data is not None, 'No element_data passed in!'
            return {'<text>': 'Madame Satan'}
        self.t.somethod = somethod
        self.makeRenderTest ("<bar>Madame Satan</bar>")
    
    def test14_renderTextDirectiveWithAssets (self):
        "Try to use the hdom:text directive containing hdom:id assets"
        self.t.manage_edit (text='<foo xmlns:hdom="uri:hiperdom" hdom:text="somethod"><bar hdom:id="bar">Madame Satan</bar></foo>')
        def somethod (bar=None):
            assert bar is not None, 'Asset lost!'
            assert bar == '<bar>Madame Satan</bar>', 'Incorrect asset passed '\
                'in: %s' % bar
            return bar
        self.t.somethod = somethod
        self.makeRenderTest ("<bar>Madame Satan</bar>")
        pass

    def test15_renderProlog(self):
        """Check for complete prolog output
        
        If a more extensive prolog is included in a XML document (including
        comments and processing instructions), HiperDom should preserve these in
        the output.
        """
        self.t.manage_edit(text='<?xml version="1.0"?>'
            '<!DOCTYPE foo SYSTEM "uri:bar/baz">'
            '<!-- This is a comment -->'
            '<?hiperdom And this is a "HiperDom" PI?>'
            '<foo/>')
        self.makeRenderTest('<!DOCTYPE foo SYSTEM "uri:bar/baz">\n'
            '<!-- This is a comment -->'
            '<?hiperdom And this is a "HiperDom" PI?>'
            '<foo/>')

    def test16_renderEmptyXHTMLElements(self):
        """Check for XHTML output, focusing on empty elements"""

        self.t.manage_edit(text='<html><br/></html>')
        self.makeRenderTest('<br />')

    def test17_renderTextExpand(self):
        """Expand the hdom:text elements in a template

        This is a sort of pre-fillin, using only the hdom:text attributes.

        """

        self.t.manage_edit(text='<foo xmlns:hdom="uri:hiperdom" '
            'hdom:text="bar"></foo>')
        self.t.bar = 'Madame Satan'
        
        try:
            t = self.t.text_expand()
        except:
            self.fail("Couldn't expand template:" + ZUnit.subTraceback())

        self.failUnless(t == "<?xml version='1.0' encoding='UTF-8'?>\n"
            "<foo xmlns:hdom='uri:hiperdom' hdom:text='bar'>Madame Satan"
            "</foo>",
            'The template was expanded incorrectly:\n         ' +
            string.join (string.split (t, '\n'), '\n         '))

    def test18_renderIncludeDirective(self):
        """Render a hdom:include directive.

        A hdom:include directive includes another HiperDom document.

        """
        self.t.manage_edit(text='<foo xmlns:hdom="uri:hiperdom">'
            '<bar hdom:include="bar"/></foo>')
        
        ctx = self.ZopeContext
        HiperDomTemplate.manage_addHiperDomTemplate(ctx, 'bar', 
            ctx.REQUEST)
        ctx.bar.manage_edit(text='<bar>Madam Satan</bar>')
        self.t.bar = ctx.bar

        self.makeRenderTest("<bar>Madam Satan</bar>")

    def test19_renderCallDirective(self):
        """Render a template that has a call directive"""
        self.t.manage_edit(text='<foo xmlns:hdom="uri:hiperdom" '
            'hdom:call="somethod"><bar hdom:dummy="1"/></foo>')

        def somethod (self):
            self.testSuccess = 1
            return '<bar/>'
        self.t.somethod = somethod

        self.makeRenderTest("<bar", invert=1)
        assert self.t.testSuccess, "Somethod not called."

def creator():
    suite = ZUnit.TestSuite()
    suite.addTest (creationTest())
    suite.addTest (ZUnit.makeSuite (objectTests))
    return suite
